from labelpandas import Client as lpClient
from tqdm import tqdm
from labelpandas.connector import get_col_names, get_table_length
import pandas

def load_local_files(client:lpClient, table:pandas.core.frame.DataFrame, file_path_column: str, verbose=False):
    """ Creates temporary URLs that can be used to create Labelbox data rows
    Args:
        client            :     Required (labelpandas.client.Client) - LabelPandas Client object
        table             :     Required (pandas.core.frame.DataFrame) - Pandas DataFrame
        file_path_column  :     Required (str) - Column with local file paths as strings
        verbose             :   Required (bool) - If True, prints details about code execution; if False, prints minimal information
    Returns:
        table with new `row_data` column which has urls to-be-used as row data in creating Labelbox data rows
    """
    existing_cols = get_col_names(table=table, extra_client=None)
    if "row_data" in existing_cols:
        print(f"Warning - column with name `row_data` already exists - renaming this column to `old_row_data` and creating a new `row_data` column with URL values to-be-used in creating Labelbox data rows")
        table = table.rename(columns={"row_data":"old_row_data"})
    url_column = []
    if verbose:
        table_length = get_table_length(table=table, extra_client=None)
        print(f"Creating URLs for {table_length} local files...")
        for file_path in tqdm(table[file_path_column]):
            url_column.append(client.lb_client.upload_file(file_path))
        print(f"Success - {table_length} URLs created in `row_data` column")
    else:
        for file_path in table[file_path_column]:
            url_column.append(client.lb_client.upload_file(file_path))            
    table["row_data"] = url_column
    return table



"""
connector.py holds the following helper functions specific to pandas DataFrames:

- get_col_names(table)                                    : Gets all column names
- get_unique_values(table, col)                           : Gets all unique values in a given column
- add_col(table, col, default)                            : Creates a column where all values equal the default value
- get_table_length(table)                                 : Gets the length of a DataFrame
- rename_col(table, col, to)                              : Renames a given column

"""
import pandas   
  
def get_col_names(table:pandas.core.frame.DataFrame, extra_client=None):
    """Grabs all column names from a Pandas DataFrame
    Args:
        table           :   Required (pandas.core.frame.DataFrame) - Pandas DataFrame
        extra_client    :   Ignore this value - necessary for other labelbase integrations
    Returns:
        List of strings corresponding to all column names
    """
    return [c for c in table.columns]

def get_unique_values(table:pandas.core.frame.DataFrame, col:str, extra_client=None):
    """Grabs all unique values from a column in a Pandas DataFrame
    Args:
        table           :   Required (pandas.core.frame.DataFrame) - Pandas DataFrame
        col             :   Required (str) - Column name
        extra_client    :   Ignore this value - necessary for other labelbase integrations
        
    Returns:
        List of strings corresponding to all unique values in a column
    """    
    return list(table[col].unique())

def add_col(table:pandas.core.frame.DataFrame, col:str, default="", extra_client=None):
    """ Adds a column of empty values to an existing Pandas DataFrame
    Args:
        table           :   Required (pandas.core.frame.DataFrame) - Pandas DataFrame
        col             :   Required (str) - Column name
        default         :   Optional - Value to insert for every row in the newly created column
        extra_client    :   Ignore this value - necessary for other labelbase integrations        
    Returns:
        Your table with a new column given the column name and default value
    """
    table[col] = default_value
    return table

def get_table_length(table:pandas.core.frame.DataFrame, extra_client=None):
    """ Tells you the size of a Pandas DataFrame
    Args:
        table           :   Required (pandas.core.frame.DataFrame) - Pandas DataFrame
        extra_client    :   Ignore this value - necessary for other labelbase integrations        
    Returns:
        The length of your table as an integer
    """  
    return len(table)



"""
uploader.py holds the function create_upload_dict() -- which creates the following style dictionary:
{
    global_key : {
        "data_row" : {}, -- This is your data row upload as a dictionary
        "dataset_id" : "" -- This is the dataset a global_key should go to (or belongs to)
        "project_id" : "", -- This batches data rows to projects, if applicable
        "annotations" : [], -- List of annotations for a given data row, if applicable
        "model_run_id" : "", -- This adds data rows to model runs, if applicable
        "predictions" : [] -- List of predictions for a given data row, if applicable
    },
    global_key : {
        "data_row" : {}, -- This is your data row upload as a dictionary
        "dataset_id" : "" -- This is the dataset a global_key should go to (or belongs to)
        "project_id" : "", -- This batches data rows to projects, if applicable
        "annotations" : [], -- List of annotations for a given data row, if applicable
        "model_run_id" : "", -- This adds data rows to model runs, if applicable
        "predictions" : [] -- List of predictions for a given data row, if applicable
    }    
}

This uniforms input formats so that they can leverage labelbase - Labelbox base code for best practices
"""

import pandas
from labelbox import Client as labelboxClient
from labelpandas.connector import get_table_length, get_unique_values
from labelbase.metadata import get_metadata_schema_to_name_key, process_metadata_value
from labelbase.ontology import get_ontology_schema_to_name_path
from labelbase.models import create_model_run_with_name
from labelbase.annotate import create_ndjsons
from concurrent.futures import ThreadPoolExecutor, as_completed
from uuid import uuid4

def create_upload_dict(client:labelboxClient, table: pandas.core.frame.DataFrame, table_dict:dict, 
                       row_data_col:str, global_key_col:str, external_id_col:str, 
                       dataset_id_col:str, dataset_id:str, project_id_col:str, project_id:str,
                       model_id_col:str, model_id:str, model_run_id_col:str, model_run_id:str,
                       metadata_index:dict, attachment_index:dict, annotation_index:dict, prediction_index:dict,
                       create_action, annotate_action, prediction_action, batch_action,
                       upload_method:str, mask_method:str, divider:str, verbose:bool, extra_client:bool=None):
    """
    Args:
        client                      :   Required (labelbox.client.Client) - Labelbox Client object        
        table                       :   Required (pandas.core.frame.DataFrame) - Pandas DataFrame                
        table_dict                  :   Required (dict) - Pandas DataFrame as dict with df.to_dict("records")
        row_data_col                :   Required (str) - Column containing asset URL or raw text
        global_key_col              :   Required (str) - Column name containing the data row global key - defaults to row data
        external_id_col             :   Required (str) - Column name containing the data row external ID - defaults to global key
        dataset_id_col              :   Required (str) - Column name containing the dataset ID to add data rows to
        dataset_id                  :   Required (str) - Labelbox dataset ID to add data rows to - only necessary if no "dataset_id" column exists            
        project_id_col              :   Required (str) - Column name containing the project ID to batch a given row to        
        project_id                  :   Required (str) - Labelbox project ID to add data rows to - only necessary if no "project_id" column exists
        model_id_col                :   Required (str) - Column name containing the model ID to add ground truth / predictions to (will create model runs)
        model_id                    :   Required (str) - Labelbox model ID to add ground truth / predictions to (will create a model run)
        model_run_id_col            :   Required (str) - Column name containing the model run ID to add ground truth / predictions to
        model_run_id                :   Required (str) - Labelbox model run ID to add ground truth / predictions to
        metadata_index              :   Required (dict) - Dictonary where {key=metadata_field_name : value=metadata_type}
        attachment_index            :   Required (dict) - Dictonary where {key=column_name : value=attachment_type}
        annotation_index            :   Required (dict) - Dictonary where {key=column_name : value=top_level_feature_name}
        prediction_index            :   Required (dict) - Dictonary where {key=column_name : value=top_level_feature_name}
        create_action               :   Required (bool) - If True, creates "data_row" key in upload_dict
        annotate_action             :   Required (str) - If any value, creates "annotations" key in the upload_dict
        prediction_action           :   Required (bool) - If True, creates "predictions" key in the upload_dict
        mask_method                 :   Optional (str) - Specifies your input mask data format
                                            - "url" means your mask is an accessible URL (must provide color)
                                            - "array" means your mask is a numpy array (must provide color)
                                            - "png" means your mask value is a png-string                 
        divider                     :   Required (str) - String delimiter for all name keys generated
        verbose                     :   Required (bool) - If True, prints details about code execution; if False, prints minimal information
        extra_client                :   Ignore this value - necessary for other labelbase integrations                
    Returns:
        - global_key_to_upload_dict - Dictionary in the above format
    """    
    # Check that global key column is entirely unique values
    table_length = get_table_length(table=table, extra_client=extra_client)
    if verbose:
        print(f'Creating upload list - {table_length} rows in Pandas DataFrame')
    # Get your global keys
    global_keys = get_unique_values(table=table, col=global_key_col, extra_client=extra_client)    
    if table_length != len(global_keys):
        print(f"Warning: Your global key column is not unique - upload will resume, only uploading 1 data row per unique global key")  
    # Get dictionaries where {key=metadata_name_key : value=metadata_schema_id}
    metadata_name_key_to_schema = get_metadata_schema_to_name_key(client=client, lb_mdo=False, divider=divider, invert=True) 
    # Get a list of project_ids
    if project_id != "":
        project_ids = [project_id]
    elif project_id_col != "":
        project_ids = get_unique_values(table=table, col=project_id_col, extra_client=extra_client)
    else:
        project_ids = []
    # Get dictionary where {key=project_id : value=ontology_index} 
    project_id_to_ontology_index = {}
    if annotate_action:
        for projectId in project_ids:
            ontology_index = get_ontology_schema_to_name_path(
                ontology=client.get_project(projectId).ontology(), 
                divider=divider, invert=True, detailed=True
            )
            ontology_index['project_type'] = str(client.get_project(projectId).media_type)
            project_id_to_ontology_index[projectId] = ontology_index
    # Get dictionary where {key=model_run_id : value=ontology_index}
    model_run_id_to_ontology_index = {}      
    # Get a dictionary where { key=model_id : value=model_run_id }    
    model_id_to_model_run_id = {}
    if prediction_action:
        if model_run_id != "": # If given a model_run_id, use it
            model_id = client.get_model_run(model_run_id).model_id
            model_id_to_model_run_id[model_id] = model_run_id
        elif model_run_id_col != "": # Else, use a model_run_id column if there is one
            model_run_ids = get_unique_values(table=table, col=model_run_id_col, extra_client=extra_client)
            for model_run_id in model_run_ids:
                model_id = client.get_model_run(model_run_id).model_id
                model_id_to_model_run_id[model_id] = model_run_id
        elif model_id != "": # Else, use a model_id keyword argument if there is one
            model = client.get_model(model_id)
            model_run = create_model_run_with_name(model=model)
            if verbose:
                print(f"Created model_run with name '{model_run.name}' for Labelbox Model with name {model.name}")
            model_id_to_model_run_id[model.uid] = model_run.uid
        elif model_id_col != "": # Else, use a model_id column if there is one
            model_ids = get_unique_values(table=table, col=model_id_col, extra_client=extra_client)
            for model_id in model_ids:
                model = client.get_model(model_id)
                model_run = create_model_run_with_name(model=model)
                if verbose:
                    print(f"Created model_run with name '{model_run.name}' for Labelbox Model with name {model.name}")                
                model_id_to_model_run_id[model.uid] = model_run.uid
        else:
            model_id_to_model_run_id = {}                                               
        for model_id in model_id_to_model_run_id.keys():
            model_run_id = model_id_to_model_run_id[model_id]
            ontology_index = get_ontology_schema_to_name_path(
                ontology=client.get_model(model_id).ontology(), divider=divider, invert=True, detailed=True
            )
            model_run_id_to_ontology_index[model_run_id] = ontology_index     
    # Multithread creating upload_dict global_key dictionaries
    upload_dict = {}
    with ThreadPoolExecutor(max_workers=8) as exc:
        futures = []
        for row_dict in table_dict:
            futures.append(exc.submit(
                create_upload, row_dict, row_data_col, global_key_col, external_id_col, 
                dataset_id_col, dataset_id, project_id_col, project_id, 
                model_id_col, model_id, model_run_id_col, model_run_id,
                metadata_index, attachment_index, annotation_index, prediction_index,
                project_id_to_ontology_index, model_run_id_to_ontology_index, model_id_to_model_run_id,
                create_action, annotate_action, prediction_action, batch_action,
                metadata_name_key_to_schema, upload_method, mask_method, divider, verbose 
            ))
        for f in as_completed(futures):
            res = f.result()
            global_key = str(res["data_row"]["global_key"])
            upload_dict[global_key] = {
                "data_row" : res["data_row"],
                "dataset_id" : res["dataset_id"],
                "project_id" : res["project_id"],
                "annotations" : res["annotations"],
                "model_run_id" : res["model_run_id"],
                "predictions" : res["predictions"]
            }
    if verbose:
        print(f'Upload generated')  
    return upload_dict

def create_upload(row_dict:dict, row_data_col:str, global_key_col:str, external_id_col:str, 
                  dataset_id_col:str, dataset_id:str, project_id_col:str, project_id:str,
                  model_id_col:str, model_id:str, model_run_id_col:str, model_run_id:str,
                  metadata_index:dict, attachment_index:dict, annotation_index:dict, prediction_index:dict,
                  project_id_to_ontology_index:dict, model_run_id_to_ontology_index:dict, model_id_to_model_run_id:dict, 
                  create_action:bool, annotate_action:bool, prediction_action:bool, batch_action:bool,
                  metadata_name_key_to_schema:dict, upload_method:str, mask_method:str, 
                  divider:str, verbose:bool):
    """ Takes a single table row as-a-dictinary and returns a dictionary where:
    {
        "data_row" : {
            "global_key" : "",                    |
            "row_data" : "",                      |
            "external_id" : "",                   | ---- This is your data row upload as a dictionary
            "metadata_fields" : [],               |
            "attachments" : []                    |
        }, -- This is your data row upload as a dictionary
        "dataset_id" : "" -- This is the dataset ID to upload this data row to, if applicable
        "project_id" : "", -- This batches data rows to projects, if applicable
        "annotations" : [], -- List of annotations for a given data row, if applicable
        "model_run_id" : "", -- This adds data rows to model runs, if applicable
        "predictions" : [] -- List of predictions for a given data row, if applicable
    }
    """
    #Remove nan values from dictionary
    nan_keys = []
    for key in row_dict.keys():
        if pandas.isna(row_dict[key]):
            nan_keys.append(key)
    for key in nan_keys:
        del row_dict[key]
    # Determine dataset ID
    if dataset_id:
        datasetId = dataset_id
    elif dataset_id_col:
        datasetId = row_dict[dataset_id_col]
    else:
        datasetId = ""                                                                                
    # Determine project ID
    if project_id:
        projectId = project_id
    elif project_id_col:
        projectId = row_dict[project_id_col]
    else:
        projectId = ""
    # Determine model run ID - check for model run first, model second  + keyword arguments first, columns second, 
    if model_run_id:
        modelRunId = model_run_id
    elif model_run_id_col: 
        modelRunId = row_dict[model_run_id_col]
    elif model_id:
        modelRunId = model_id_to_model_run_id[model_id]                                                                               
    elif model_id_col:
        model_id = row_dict[model_id_col]                                                                 
        modelRunId = model_id_to_model_run_id[model_id]
    else:
        modelRunId = ""   
    # Create a base data row dictionary     
    data_row = {}
    if create_action or batch_action:
        data_row["row_data"] = row_dict[row_data_col]
        if len(row_dict[global_key_col]) <= 200:
            data_row["global_key"] = row_dict[global_key_col]
        else:
            if verbose:
                print("Global key too long (>200 characters). Replacing with randomly generated global key.")
            data_row["global_key"] = str(uuid4())
        if external_id_col in row_dict.keys():
            data_row["external_id"] = row_dict[external_id_col]
        # Create a list of metadata for a data row    
        metadata_fields = [{"schema_id" : metadata_name_key_to_schema['lb_integration_source'], "value" : "LabelPandas"}]
        if metadata_index:
            for metadata_field_name in metadata_index.keys():
                metadata_type = metadata_index[metadata_field_name]
                column_name = f"metadata{divider}{metadata_type}{divider}{metadata_field_name}"
                if column_name in row_dict.keys():
                    input_metadata = process_metadata_value(
                        metadata_value=row_dict[column_name], metadata_type=metadata_type, 
                        parent_name=metadata_field_name, metadata_name_key_to_schema=metadata_name_key_to_schema, divider=divider
                    )            
                    if input_metadata:
                        metadata_fields.append({"schema_id" : metadata_name_key_to_schema[metadata_field_name], "value" : input_metadata})
                    else:
                        continue        
        data_row["metadata_fields"] = metadata_fields  
        # Create a list of attachments for a data row
        if attachment_index:
            data_row["attachments"] = [{"type" : attachment_index[column_name], "value" : row_dict[column_name]} for column_name in attachment_index]
    # Create a list of annotation ndjsons for a data row (does not include data row ID since data row has not been created)
    annotations = []        
    if annotate_action:
        ontology_index = project_id_to_ontology_index[projectId]
        for column_name in annotation_index.keys():
            annotations.extend(
                create_ndjsons(
                    top_level_name=annotation_index[column_name],
                    annotation_inputs=row_dict[column_name],
                    ontology_index=ontology_index,
                    confidence=False,
                    mask_method=mask_method,
                    divider=divider
                )
            )
    # Create a list of predictoin ndjsons for a data row (does not include data row ID since data row has not been created)
    predictions = []        
    if prediction_action:
        ontology_index = model_run_id_to_ontology_index[modelRunId]
        for column_name in prediction_index.keys():
            predictions.extend(
                create_ndjsons(
                    top_level_name=prediction_index[column_name],
                    annotation_inputs=row_dict[column_name],
                    ontology_index=ontology_index,
                    confidence=True,
                    mask_method=mask_method,
                    divider=divider
                )
            )                                                                                
    return {
        "data_row" : data_row,
        "project_id" : projectId,
        "dataset_id" : datasetId,
        "annotations" : annotations,
        "model_run_id" : modelRunId,
        "predictions" : predictions
    }
    



from labelbox import Client as labelboxClient
from labelpandas import uploader, connector
from labelbase.downloader import *
from labelbase.uploader import *
from labelbase.connector import *
import pandas as pd


class Client():
    """
    Args:
        lb_api_key                  :   Required: labelbox.client.Client()
    Attributes:
        lb_client                   :   labelbox.Client object
    Key Functions:
        create_data_rows_from_table :   Creates Labelbox data rows (and metadata) given a Pandas DataFrame an an existing Labelbox Dataset
    """       
    def __init__(
        self,           
        lb_api_key:str=None,
        lb_endpoint='https://api.labelbox.com/graphql', 
        lb_enable_experimental=False, 
        lb_app_url="https://app.labelbox.com"):

        self.lb_client = labelboxClient(lb_api_key, endpoint=lb_endpoint, enable_experimental=lb_enable_experimental, app_url=lb_app_url)
           
    def export_to_table(
        self, project,
        include_metadata:bool=False, include_performance:bool=False, include_agreement:bool=False,
        verbose:bool=False, mask_method:str="png", divider="///"):
        """ Creates a Pandas DataFrame given a Labelbox Projet ID
        Args:
            project                 :   Required (str / lablebox.Project) - Labelbox Project ID or lablebox.Project object to export labels from
            include_metadata        :   Optional (bool) - If included, exports metadata fields
            include_performance     :   Optional (bool) - If included, exports labeling performance
            include_agreement       :   Optional (bool) - If included, exports consensus scores       
            verbose                 :   Optional (bool) - If True, prints details about code execution; if False, prints minimal information
            mask_method             :   Optional (str) - Specifies your desired mask data format
                                            - "url" leaves masks as-is
                                            - "array" converts URLs to numpy arrays
                                            - "png" converts URLs to png byte strings             
            divider                 :   Optional (str) - String delimiter for schema name keys and suffix added to duplocate global keys 
        """
        # Create a list of dictionaries where { key=column_name : value=row_value } using labelbase.downloader.export_and_flatten_labels
        flattened_labels_dict = export_and_flatten_labels(
            client=self.lb_client, project=project, include_metadata=include_metadata, 
            include_performance=include_performance, include_agreement=include_agreement,
            mask_method=mask_method, verbose=verbose, divider=divider
        )
        # Convert to a Pandas DataFrame
        table = pd.DataFrame.from_dict(flattened_labels_dict)
        if verbose:
            print(f"Success: DataFrame generated")
        return table 

    def create_data_rows_from_table(
        self, table:pd.core.frame.DataFrame, dataset_id:str="", project_id:str="", model_id:str="", model_run_id:str="", priority:int=5, 
        upload_method:str="", skip_duplicates:bool=False, mask_method:str="png", verbose:bool=False, divider="///"):
        """ Performs the following actions if proper information is provided:
                - creates data rows with metadata and attachments *
                - batches data rows to projects * **
                - uploads annotations as pre-labels or submitted labels * **
                - sends submitted labels to model runs as ground truth labels * ** ***
                - uploads predictions to a model run * ***
            * Requires a row_data AND dataset_id OR dataset_id_col
            ** Requires project_id OR project_id_col
            *** Requires model_id OR model_id_col OR model_run_id OR model_run_id_col
        Args:
            table               :   Required (pandas.core.frame.DataFrame) - Pandas DataFrame    
            dataset_id          :   Required (str) - Labelbox dataset ID to add data rows to - only necessary if no "dataset_id" column exists            
            project_id          :   Required (str) - Labelbox project ID to add data rows (and predictions) to - only necessary if no "project_id" column exists
            model_id            :   Required (str) - Labelbox model ID to add data rows, annotations and/or predictions to - will create a model run
            model_run_id        :   Required (str) - Labelbox model run ID to add data rows, annotations and/or predictions to - will use an existing model run
            priority            :   Optional (int) - Between 1 and 5, what priority to give to data row batches sent to projects                             
            upload_method       :   Optional (str) - Must be one of the following:
                                            "mal" - Project pre-labels (requires project ID)
                                            "import" - Project submitted labels (requires project ID)
                                            "ground-truth" - Project submitted labels AND ground truth submitted labels (requires project ID AND model / model run ID)
            skip_duplicates     :   Optional (bool) - Determines how to handle if a global key to-be-uploaded is already in use
                                        If True, will skip duplicate global_keys and not upload them
                                        If False, will generate a unique global_key with a suffix {divider} + "1", "2" and so on
            mask_method         :   Optional (str) - Specifies your input mask data format
                                        - "url" means your mask is an accessible URL (must provide color)
                                        - "array" means your mask is a numpy array (must provide color)
                                        - "png" means your mask value is a png-string                                           
            verbose             :   Optional (bool) - If True, prints details about code execution; if False, prints minimal information               
            divider             :   Optional (str) - String delimiter for schema name keys and suffix added to duplocate global keys
        Returns:
            Results from all performed actions in a dictionary - if an expected action has no results, it was not performed
        """
        # Confirm the existience of specific columns and construct reference indexes for different kinds of data to-be-uploaded using labelbase.connector.validate_columns
            # row_data_col                :   Column representing asset URL, raw text, or path to local asset file
            # global_key_col              :   Defaults to row_data_col
            # external_id_col             :   Defaults to global_key_col
            # project_id_col              :   Either `project_id` or "" - overridden by arg for project_id        
            # dataset_id_col              :   Either `dataset_id` or "" - overridden by arg for dataset_id
            # model_id_col                :   Either `model_id` or "" - overridden by arg for model_id and any args for model_run_id
            # model_run_id_col            :   Either `model_run_id` or "" - overridden by arg for model_run_id        
            # metadata_index              :   Dictonary where {key=metadata_field_name : value=metadata_type} or {} if not uploading metadata
            # attachment_index            :   Dictonary where {key=column_name : value=attachment_type} or {} if not uploading attachments
            # annotation_index            :   Dictonary where {key=column_name : value=top_level_class_name} or {} if not uploading annotations
            # prediction_index            :   Dictonary where {key=column_name : value=top_level_class_name} or {}  if not uploading predictions
        x = validate_columns(
            client=self.lb_client, table=table,
            get_columns_function=connector.get_col_names,
            get_unique_values_function=connector.get_unique_values,
            divider=divider, verbose=verbose, extra_client=None
        )
        
        # Determine the actions we're taking in this run using labelbase.connector.determine_actions
        actions = determine_actions(
            row_data_col=x["row_data_col"], dataset_id=dataset_id, dataset_id_col=x["dataset_id_col"], project_id=project_id, 
            project_id_col=x["project_id_col"], model_id=model_id, model_id_col=x["model_id_col"], model_run_id=model_run_id, 
            model_run_id_col=x["model_run_id_col"], upload_method=upload_method, metadata_index=x["metadata_index"], 
            attachment_index=x["attachment_index"], annotation_index=x["annotation_index"], prediction_index=x["prediction_index"]
        )
        
        # Since this function's primary function is to create data rows, we raise an error if the create_action wasn't triggered
        if not actions["create"]:
            raise ValueError(f"No `dataset_id` argument or `dataset_id` column was provided in table to create data rows")
        
        # Iterating over your pandas DataFrame is faster once converted to a list of dictionaries where {keys=column_names : value=row_column_values}
        table_dict = table.to_dict('records')

        # Create an upload dictionary where:
        # {
            # global_key : {
                # "data_row" : {}, -- This is your data row upload as a dictionary
                # "dataset_id" : "", -- This is the dataset a global_key should go to (or belongs to)
                # "project_id" : "", -- This batches data rows to projects, if applicable
                # "annotations" : [], -- List of annotations for a given data row, if applicable
                # "model_run_id" : "", -- This adds data rows to model runs, if applicable
                # "predictions" : [] -- List of predictions for a given data row, if applicable
            # }
        # }
        # This uniforms the upload to use labelbase - Labelbox base code for best practices
        upload_dict = uploader.create_upload_dict( # Using labelpandas.uploader.create_upload_dict
            client=self.lb_client, table=table, table_dict=table_dict, 
            row_data_col=x["row_data_col"], global_key_col=x["global_key_col"], external_id_col=x["external_id_col"], 
            dataset_id_col=x["dataset_id_col"], dataset_id=dataset_id, 
            project_id_col=x["project_id_col"], project_id=project_id,
            model_id_col=x["model_id_col"], model_id=model_id, 
            model_run_id_col=x["model_run_id_col"], model_run_id=model_run_id,
            metadata_index=x["metadata_index"], attachment_index=x["attachment_index"], 
            annotation_index=x["annotation_index"], prediction_index=x["prediction_index"], batch_action=actions["batch"],
            create_action=actions["create"], annotate_action=actions["annotate"], prediction_action=actions["predictions"],
            upload_method=upload_method, mask_method=mask_method, divider=divider, verbose=verbose
        )
                
        # Upload your data rows to Labelbox, updating upload_dict if global keys are modified during upload using labelbase.uploader.batch_create_data_rows
        data_row_upload_results, upload_dict = batch_create_data_rows(
            client=self.lb_client, upload_dict=upload_dict, 
            skip_duplicates=skip_duplicates, divider=divider, verbose=verbose
        )
        
        # Bath to project attempt using labelbase.uploader.batch_rows_to_project
        if actions["batch"]:         
            batch_to_project_results = batch_rows_to_project(
                client=self.lb_client, upload_dict=upload_dict, 
                priority=priority, verbose=verbose
            )                
        else:
            batch_to_project_results = []
            
        # If performing actions that require data row IDs, we pull them here using labelbase.uploader.create_global_key_to_data_row_id_dict
        if actions["annotate"] or actions["predictions"]:
            global_key_to_data_row_id = create_global_key_to_data_row_id_dict(
                client=self.lb_client, global_keys=list(upload_dict.keys())
            )            
        
        # Annotation upload attempt using labelbase.uploader.batch_upload_annotations
        if actions["annotate"]:
            annotation_upload_results = batch_upload_annotations(
                client=self.lb_client, upload_dict=upload_dict, 
                global_key_to_data_row_id=global_key_to_data_row_id,
                how=actions["annotate"], verbose=verbose
            )
        else:
            annotation_upload_results = []

        # If model run upload attempt labelbase.uploader.batch_add_ground_truth_to_model_run
        if actions["annotate"] == "ground-truth":
            ground_truth_upload_results = batch_add_ground_truth_to_model_run(
                client=self.lb_client, upload_dict=upload_dict
            )
        else:
            ground_truth_upload_results = []
            
        # Prediction upload attempt using labelbase.uploader.batch_upload_predictions
        if actions["predictions"]:
            # If data rows aren't in the model run yet, add them with labelbase.uploader.batch_add_data_rows_to_model_run
            if not ground_truth_upload_results:
                data_row_to_model_run_results = batch_add_data_rows_to_model_run(
                    client=self.lb_client, upload_dict=upload_dict
                )
            prediction_upload_results = batch_upload_predictions(
                client=self.lb_client, upload_dict=upload_dict,
                global_key_to_data_row_id=global_key_to_data_row_id
            )
        else:
            data_row_to_model_run_results = []
            prediction_upload_results = []              

        return_payload = {
            "data_row_upload_results" : data_row_upload_results, 
            "batch_to_project_results" : batch_to_project_results,
            "annotation_upload_results" : annotation_upload_results,
            "ground_truth_upload_results" : ground_truth_upload_results,
            "data_row_to_model_run_results" : data_row_to_model_run_results,
            "prediction_upload_results" : prediction_upload_results
        }             

        return_payload = {k: v for k, v in return_payload.items() if v}
            
        return return_payload
    
    ## to do - add support for upserting metadata and upserting attachments
    # def upsert_data_rows_from_table(batch_data_rows:bool=False, anootation_method:str="", upsert_metadata:bool=False, upsert_attachments:bool=False):
    #     """
    #     Args:
    #         batch_data_rows     :   Optional (bool) - If True, will batch data rows to projects if not already in projects
    #         anootation_method   :   Optional (bool) -  Must be one of the following annotation import methods:
    #                                         "" - Does nothing with annotations
    #                                         "mal" - Uploaded to project as pre-labels
    #                                         "import" - Uploaded to project assubmitted labels
    #                                         "ground-truth" - Project submitted labels AND/OR ground truth submitted labels (will check existing project labels first)
    #         upsert_metadata     :   True if upserting metadata, False if not
    #         upsert_attachments  :   True if upserting attachments, False if not
    #     Returns:
    #         Results from all performed actions in a dictionary - if an expected action has no results, it was not performed
    #     """    
    def upsert_data_rows_from_table(self, table:pd.core.frame.DataFrame, dataset_id:str="", project_id:str="", model_id:str="", upload_method:str="", mask_method:str="png", priority:int=5, model_run_id:str="", batch_data_rows:bool=False, verbose:bool=False, divider:str="///"):
        """ Performs the following actions if proper information is provided:
                - batches data rows to projects (if batch_data_rows == True) * **
                - uploads annotations as pre-labels or submitted labels * **
                - sends submitted labels to model runs as ground truth labels * ** ***
                - uploads predictions to a model run * ***
            * Requires a global_key AND dataset_id OR dataset_id_col
            ** Requires project_id OR project_id_col
            *** Requires model_id OR model_id_col OR model_run_id OR model_run_id_col      
        Args:
            batch_data_rows     :   Optional (bool) - If True, will batch data rows to projects if not already in projects
            anootation_method   :   Optional (bool) -  Must be one of the following annotation import methods:
                                            "" - Does nothing with annotations
                                            "mal" - Uploaded to project as pre-labels
                                            "import" - Uploaded to project assubmitted labels
                                            "ground-truth" - Project submitted labels AND/OR ground truth submitted labels (will check existing project labels first)
            upsert_metadata     :   True if upserting metadata, False if not
            upsert_attachments  :   True if upserting attachments, False if not
        Returns:
            Results from all performed actions in a dictionary - if an expected action has no results, it was not performed
        """
        # Confirm the existience of specific columns and construct reference indexes for different kinds of data to-be-uploaded using labelbase.connector.validate_columns
            # row_data_col                :   Column representing asset URL, raw text, or path to local asset file
            # global_key_col              :   Defaults to row_data_col
            # external_id_col             :   Defaults to global_key_col
            # project_id_col              :   Either `project_id` or "" - overridden by arg for project_id        
            # dataset_id_col              :   Either `dataset_id` or "" - overridden by arg for dataset_id
            # model_id_col                :   Either `model_id` or "" - overridden by arg for model_id and any args for model_run_id
            # model_run_id_col            :   Either `model_run_id` or "" - overridden by arg for model_run_id        
            # metadata_index              :   Dictonary where {key=metadata_field_name : value=metadata_type} or {} if not uploading metadata
            # attachment_index            :   Dictonary where {key=column_name : value=attachment_type} or {} if not uploading attachments
            # annotation_index            :   Dictonary where {key=column_name : value=top_level_class_name} or {} if not uploading annotations
            # prediction_index            :   Dictonary where {key=column_name : value=top_level_class_name} or {}  if not uploading predictions
        x = validate_columns(
            client=self.lb_client, table=table,
            get_columns_function=connector.get_col_names,
            get_unique_values_function=connector.get_unique_values,
            divider=divider, verbose=verbose, extra_client=None
        )
        
        # Determine the actions we're taking in this run using labelbase.connector.determine_actions
        actions = determine_actions(
            row_data_col=x["row_data_col"], dataset_id=dataset_id, dataset_id_col=x["dataset_id_col"], project_id=project_id, 
            project_id_col=x["project_id_col"], model_id=model_id, model_id_col=x["model_id_col"], model_run_id=model_run_id, 
            model_run_id_col=x["model_run_id_col"], upload_method=upload_method, metadata_index=x["metadata_index"], 
            attachment_index=x["attachment_index"], annotation_index=x["annotation_index"], prediction_index=x["prediction_index"]
        )

        # Iterating over your pandas DataFrame is faster once converted to a list of dictionaries where {keys=column_names : value=row_column_values}
        table_dict = table.to_dict('records')                
        
        # Create an upload dictionary where:
        # {
            # global_key : {
                # "project_id" : "", -- This batches data rows to projects, if applicable
                # "annotations" : [], -- List of annotations for a given data row, if applicable
                # "model_run_id" : "", -- This adds data rows to model runs, if applicable
                # "predictions" : [] -- List of predictions for a given data row, if applicable
            # }
        # }
        # This uniforms the upload to use labelbase - Labelbox base code for best practices
        upload_dict = uploader.create_upload_dict( # Using labelpandas.uploader.create_upload_dict
            client=self.lb_client, table=table, table_dict=table_dict, 
            row_data_col=x["row_data_col"], global_key_col=x["global_key_col"], external_id_col=x["external_id_col"], 
            dataset_id_col=x["dataset_id_col"], dataset_id=dataset_id, 
            project_id_col=x["project_id_col"], project_id=project_id,
            model_id_col=x["model_id_col"], model_id=model_id, 
            model_run_id_col=x["model_run_id_col"], model_run_id=model_run_id,
            metadata_index=x["metadata_index"], attachment_index=x["attachment_index"], 
            annotation_index=x["annotation_index"], prediction_index=x["prediction_index"], 
            create_action=actions["create"], annotate_action=actions["annotate"], prediction_action=actions["predictions"], batch_action=actions["batch"],
            upload_method=upload_method, mask_method=mask_method, divider=divider, verbose=verbose
        )

        # Bath to project attempt using labelbase.uploader.batch_rows_to_project
        if actions["batch"] and batch_data_rows:         
            batch_to_project_results = batch_rows_to_project(
                client=self.lb_client, upload_dict=upload_dict, priority=priority, verbose=verbose
            )                
        else:
            batch_to_project_results = []

        # Annotation upload attempt using labelbase.uploader.batch_upload_annotations
        if actions["annotate"]:
            annotation_upload_results = batch_upload_annotations(
                client=self.lb_client, upload_dict=upload_dict, how=actions["annotate"], verbose=verbose
            )
        else:
            annotation_upload_results = []

        # If model run upload attempt labelbase.uploader.batch_add_ground_truth_to_model_run
        if actions["annotate"] == "ground-truth":
            ground_truth_upload_results = batch_add_ground_truth_to_model_run(
                client=self.lb_client, upload_dict=upload_dict
            )
        else:
            ground_truth_upload_results = []
            
        # Prediction upload attempt using labelbase.uploader.batch_upload_predictions
        if actions["predictions"]:
            # If data rows aren't in the model run yet, add them with labelbase.uploader.batch_add_data_rows_to_model_run
            if not ground_truth_upload_results:
                data_row_to_model_run_results = batch_add_data_rows_to_model_run(
                    client=self.lb_client, upload_dict=upload_dict                    
                )
            prediction_upload_results = batch_upload_predictions(
                client=self.lb_client, upload_dict=upload_dict
            )
        else:
            data_row_to_model_run_results = []
            prediction_upload_results = []  

        return_payload = {
            "batch_to_project_results" : batch_to_project_results,
            "annotation_upload_results" : annotation_upload_results,
            "ground_truth_upload_results" : ground_truth_upload_results,
            "data_row_to_model_run_results" : data_row_to_model_run_results,
            "prediction_upload_results" : prediction_upload_results
        }             

        return_payload = {k: v for k, v in return_payload.items() if v}
            
        return return_payload        
    
